import os
import sys

import dnf

from distutils.version import LooseVersion
from pdc_client import PDCClient


FALLBACK_STREAM = 'master'
STREAM = 'f26'
REPO_F26 = "http://ftp.fi.muni.cz/pub/linux/fedora/linux/development/26/Everything/x86_64/os/"
REPO_F26_SOURCE = "http://ftp.fi.muni.cz/pub/linux/fedora/linux/development/26/Everything/source/tree/"
ARCH = 'x86_64'

# TODO: add ability to solve more packages in one run
# TODO: solve architecture! - as option?


p = None


def get_pkgs_source_rpm_name(pkg):
    """
    pkg: dnf.Package

    returns name of source RPM used to build the provided rpm
    """
    if LooseVersion(dnf.__version__) < LooseVersion("2.0.0"):
        source_rpm = pkg.sourcerpm
        source_rpm_name = source_rpm.rsplit('-', 2)[0]
    else:
        source_rpm_name = pkg.source_name
    return source_rpm_name


def get_pdc_client():
    """ cache PDCClient instance """
    global p
    if p is None:
        p = PDCClient('https://pdc.fedoraproject.org/rest_api/v1/', develop=True, ssl_verify=True)
    return p


class ModuleDepsDiffer(object):

    def __init__(self, pkgs):
        self.p = get_pdc_client()
        self.repos = dict()
        self.repo_provides = dict()
        # TODO: get this from args
        self.packages = list(pkgs)
        self.build_req = set()
        self.runtime_req = set()
        self.build_caps_classified = dict()
        self.build_caps_classified['result'] = set()
        self.runtime_caps_classified = dict()
        self.runtime_caps_classified['result'] = set()
        self.repo_bases = {}
        self.module_id_to_name = {}

    def get_module_name(self, koji_tag_name):
        return self.module_id_to_name.get(koji_tag_name, koji_tag_name)

    def obtain_module_names(self):
        j = self.p['unreleasedvariants'](
            variant_type="module",
            active=True,
            page_size=-1
            # variant_version=STREAM,
        )
        for module in j:
            self.module_id_to_name[module["koji_tag"]] = "{}:{}".format(module["variant_id"], module["variant_version"])

    @staticmethod
    def get_base_from_repo(reponame, repourl):
        base = dnf.Base()
        if LooseVersion(dnf.__version__) < LooseVersion("2.0.0"):
            repo = dnf.repo.Repo(reponame, base.conf.cachedir)
        else:
            repo = dnf.repo.Repo(reponame, base.conf)
        repo.baseurl = repourl
        repo.load()
        repo.enable()
        base.repos.add(repo)
        base.fill_sack(load_available_repos=True, load_system_repo=False)
        return base

    def _init_repo_bases(self):
        """
        version 1 is no longer usable since koji no longer creates repos for
        modules this version iterates over repos present in
        ~/modulebuild/cache/koji_tags, this means that you should do `mbs-build
        local` to populate the directory

        Once we have real compose of boltron, we might utilize it here

        initialize repo objects for dnf to kick off queries
        """
        print("Getting module information from mbs cache." +
              " Please run mbs-build local with dependency modules" +
              " you are interested in before running this script.")
        cache_path = os.path.expanduser("~/modulebuild/cache/koji_tags")
        for module_name in os.listdir(cache_path):
            # for now ignore all what is not in f26 stream, remove when situation changes
            if not self.module_id_to_name[module_name].endswith(':f26'):
                continue
            if module_name != "f26-modularity":  # bootstrap contains almost everything, so let's ignore it
                self.repos[module_name] = "file://{}".format(os.path.join(cache_path, module_name))
        for reponame, repourl in self.repos.items():
            base = ModuleDepsDiffer.get_base_from_repo(reponame, repourl)
            self.repo_bases[reponame] = base

    def get_package_requires(self):
        # runtime
        pkgs = []
        base = ModuleDepsDiffer.get_base_from_repo('f26', REPO_F26)
        for pkg in self.packages:
            filter_result = base.sack.query().filter(name=pkg, arch=['noarch', ARCH], latest=True)
            ModuleDepsDiffer.package_unique(filter_result)
            pkgs.append(filter_result[0])
        for pkg in pkgs:
            requires = getattr(pkg, 'requires')
            for q in requires:
                self.runtime_req.add(str(q))

        # build
        source_pkgs = []
        base = ModuleDepsDiffer.get_base_from_repo('f26-source', REPO_F26_SOURCE)
        for pkg in pkgs:
            source_rpm_name = get_pkgs_source_rpm_name(pkg)
            filter_result = base.sack.query().filter(name=source_rpm_name)
            ModuleDepsDiffer.package_unique(filter_result)
            source_pkgs.append(filter_result[0])
        for pkg in source_pkgs:
            requires = getattr(pkg, 'requires')
            for q in requires:
                self.build_req.add(str(q))

    # Igor suggests to do complete dependency solving here:
    #   base.install('package'), installroot=, make sure the right repos are enabled
    #   base.resolve()
    #   and then analyze base.transaction.install_set
    # weak deps will get resolved correctly with this approach
    def classify_caps(self, caps_classified, requirements):
        for cap in requirements:
            found = False
            for reponame, base in self.repo_bases.items():
                q = base.sack.query()
                pkg = q.filter(provides=[cap], arch=['noarch', ARCH], latest=True)
                if pkg:
                    found = True
                    caps_classified.setdefault(reponame, set())
                    caps_classified[reponame].add(cap)
                    # don't break here in case one cap is present in multiple repos
            if not found:
                caps_classified['result'].add(cap)

    @staticmethod
    def package_unique(result):
        if len(result) > 1:
            raise ValueError('Name of package is not unique: ' + str(result.result))
        if len(result) == 0:
            raise ValueError('No package found in repo')

    @staticmethod
    def whatprovides(caps):
        result = set()
        base = ModuleDepsDiffer.get_base_from_repo('f26', REPO_F26)
        q = base.sack.query()
        for cap in caps:
            pkg = q.filter(provides__glob=[cap], arch=['noarch', ARCH], latest=True)
            if not pkg:
                pkg = q.filter(file__glob=cap)
                if len(pkg) == 0:
                    raise ValueError('No package provides capability \'' + cap + '\'')
            source_rpm_name = get_pkgs_source_rpm_name(pkg.result[0])
            result.add(source_rpm_name)
        return result

    def _dump_caps_result(self, items):
        for repo, caps in items:
            if not caps:
                continue
            if repo == 'result':
                print("Capabilities which aren't provided by any module:")
            else:
                print("Components which are part of " +
                      self.get_module_name(repo) +
                      " module:")
            for cap in sorted(caps):
                print(cap)
            print('')

    def dump_caps_result(self):
        print('BUILD requirements:')
        print('-------------------')
        self._dump_caps_result(self.build_caps_classified.items())
        print('\n')
        print('RUNTIME requirements:')
        print('---------------------')
        self._dump_caps_result(self.runtime_caps_classified.items())

    def _dump_pkgs_result(self, items):
        for repo, caps in items:
            if not caps:
                continue
            if repo == 'result':
                print("Components which aren't provided by any module:")
            else:
                print("Components which are part of " +
                      self.get_module_name(repo) +
                      " module:")
            for cap in sorted(ModuleDepsDiffer.whatprovides(caps)):
                print(cap)
            print('')

    def dump_pkgs_result(self):
        print('BUILD dependencies:')
        print('-------------------')
        self._dump_pkgs_result(self.build_caps_classified.items())
        print('\n')
        print('RUNTIME dependencies:')
        print('---------------------')
        self._dump_pkgs_result(self.runtime_caps_classified.items())

    def run(self):
        self.obtain_module_names()
        self._init_repo_bases()
        self.get_package_requires()
        self.classify_caps(self.build_caps_classified, self.build_req)
        self.classify_caps(self.runtime_caps_classified, self.runtime_req)
