from __future__ import absolute_import

import modulemd
import dnf
from .module_deps_differ import ModuleDepsDiffer


class ModuleGenerator(object):

    def __init__(self, pkgs):
        self.pkgs = pkgs
        self.pkg = None
        self.mmd = modulemd.ModuleMetadata()
        self.build_deps = set()
        self.run_deps = set()
        self.differ = ModuleDepsDiffer(pkgs)

    def _save_module_md(self):
        """
        Function saves modulemd file to the current directory
        based on argument name
        :return:
        """

        if len(self.pkgs) == 1:
            file_name = self.pkgs[0] + '.yaml'
        else:
            file_name = "modulemd-output.yaml"
        self.mmd.dump(file_name)
        print('Modulemd file is generated here ./%s' % file_name)
        return True

    def _update_module_md(self):
        """
        Function updates modulemd file with dependencies
        are information taken from SPEC file.
        :return:
        """
        self.mmd.add_module_license("MIT")

        if len(self.pkgs) == 1:
            self.mmd.summary = str(self.pkg.summary)
            self.mmd.description = str(self.pkg.description)

            # Default license for the module metadata, same as default Fedora
            # content license.

            self.mmd.add_content_license(str(self.pkg.license))

        for pkg in self.pkgs:
            self.mmd.api.add_rpm(pkg)
            self.mmd.components.add_rpm(ModuleDepsDiffer.whatprovides([pkg]).pop(), "Package in api", buildorder=self._get_build_order(pkg))

        for pkg in (self.build_deps - self.mmd.api.rpms - self.run_deps):
            self.mmd.filter.add_rpm(pkg)

        for pkg in self.build_deps.intersection(self.run_deps):
            self.mmd.components.add_rpm(pkg, "Build and runtime dependency.", buildorder=self._get_build_order(pkg))

        for pkg in (self.build_deps - self.run_deps):
            self.mmd.components.add_rpm(pkg, "Build dependency.", buildorder=self._get_build_order(pkg))

        for pkg in (self.run_deps - self.build_deps):
            self.mmd.components.add_rpm(pkg, "Runtime dependency.", buildorder=self._get_build_order(pkg))

        for mod, caps in self.differ.build_caps_classified.items():
            if mod == 'result':
                continue
            if caps:
                name, stream = self.differ.get_module_name(mod).split(':')
                self.mmd.add_buildrequires(name, stream)

        for mod, caps in self.differ.runtime_caps_classified.items():
            if mod == 'result':
                continue
            if caps:
                name, stream = self.differ.get_module_name(mod).split(':')
                self.mmd.add_requires(name, stream)

    def _get_build_order(self, pkg):
        if pkg in self.mmd.api.rpms:
            return 10
        else:
            return 0

    def _get_pkg_info(self):
        """
            Function loads package from dnf
            :return:
            """
        b = dnf.Base()
        b.read_all_repos()
        b.fill_sack()

        q = b.sack.query().filter(name=self.pkgs, reponame='fedora', latest=True)

        if len(q) > 1:
            raise ValueError('Name of package is not unique')
        if len(q) == 0:
            raise ValueError('No package found in repo')
        self.pkg = q[0]

    def _get_dependencies(self):
        """
            Function gets build and runtime dependencies of package
            :return:
            """
        self.differ.run()
        self.build_deps = ModuleDepsDiffer.whatprovides(self.differ.build_caps_classified['result'])
        self.run_deps = ModuleDepsDiffer.whatprovides(self.differ.runtime_caps_classified['result'])

    def run(self):
        if len(self.pkgs) == 1:
            self._get_pkg_info()
        self._get_dependencies()
        self._update_module_md()
        self._save_module_md()
