from __future__ import absolute_import, print_function

import os
import ast
import yaml
import tempfile
import shutil
import re
import shlex

from dockerfile_parse import DockerfileParser

# Dockerfile path
DOCKERFILE = "Dockerfile"

EXPOSE = "EXPOSE"
VOLUME = "VOLUME"
LABEL = "LABEL"
ENV = "ENV"
PORTS = "PORTS"

# OpenShift template
OPENSHIFT_TEMPLATE = "openshift-template.yml"


def get_string(value):
    return ast.literal_eval(value)


class OpenShiftTemplateGenerator(object):
    """
    Class generates an OpenShift template
    It requires openshift-template.yml file.
    """

    dockerfile = None
    oc_template = None
    docker_dict = {}

    def __init__(self, args=None, dir_name=None):
        if dir_name is None:
            self.dir = os.getcwd()
        else:
            self.dir = dir_name
        self.docker_image = args.image
        if args.dockerfile is None:
            self.dockerfile = 'Dockerfile'
        else:
            self.dockerfile = os.path.join(self.dir, args.dockerfile)
        self.docker_dict = {}

    def _exist_docker_file(self):
        """
        Function checks if docker file exists
        :return: True if exists
        """
        if not os.path.exists(self.dockerfile):
            print("Dockerfile has to exists in the %s directory." % self.dir)
            return False
        return True

    def _exist_openshift_template(self):
        """
        Function checks if openshift template exists
        :return: True if exists
        """
        if self.oc_template is None:
            print("%s has to exists in the %s directory." % (OPENSHIFT_TEMPLATE, self.dir))
            return False
        return True

    def _get_openshift_template(self):
        """
        Function sets openshift template.
        """
        for f in os.listdir(self.dir):
            if os.path.isdir(os.path.join(self.dir, f)):
                continue
            file_name = os.path.join(self.dir, f)
            if f == OPENSHIFT_TEMPLATE:
                self.oc_template = file_name

    def _get_expose(self, value):
        """Function returns exposes as field"""
        return value.split()

    def _get_env(self, value):
        """Function gets env as field"""
        return shlex.split(value)

    def _get_volume(self, value):
        """Function evaluates a value and returns as string."""
        return get_string(value)

    def _get_label(self, value):
        """
        Function returns label from Docker file
        except INSTALL, UNINSTALL and RUN label used by atomic.
        :param value: row from Dockerfile
        :return: label_dict
        """
        untracked_values = ['INSTALL', 'UNINSTALL', 'RUN']
        if [f for f in untracked_values if value.startswith(f)]:
            return None
        labels = re.sub('\s\s+', ';', value).split(';')
        labels = [l.replace('"', '') for l in labels]
        label_dict = {}
        for l in labels:
            if len(l.split('=')) == 2:
                label_dict[l.split('=')[0]] = l.split('=')[1]
            elif re.match('maintainer', l, re.I):
                label_dict['maintainer'] = l.split(' ',1)[1]
            else:
                raise ValueError("Unrecogised label: ", l)
        return label_dict

    def _get_docker_tags(self):
        """
        Function analyses dockerfile and extracts
        ENV, VOLUME, EXPOSE and LABEL directives.
        """
        if not self._exist_docker_file():
            return
        tmp_dir = tempfile.mkdtemp()
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)
        os.makedirs(tmp_dir)
        shutil.copyfile(self.dockerfile, os.path.join(tmp_dir, "Dockerfile"))
        dfp = DockerfileParser(path=tmp_dir)
        inst = "instruction"
        allowed_tags = [ENV, EXPOSE, VOLUME, LABEL]
        functions = {ENV: self._get_env,
                     EXPOSE: self._get_expose,
                     VOLUME: self._get_volume,
                     LABEL: self._get_label}

        for struct in dfp.structure:
            key = struct[inst]
            val = struct["value"]
            if key in allowed_tags:
                if key == LABEL:
                    if key not in self.docker_dict:
                        self.docker_dict[key] = {}
                    value = functions[key](val)
                    if value is not None:
                        self.docker_dict[key].update(value)
                else:
                    if key not in self.docker_dict:
                        self.docker_dict[key] = []
                    ret_val = functions[key](val)
                    for v in ret_val:
                        if v not in self.docker_dict[key]:
                            self.docker_dict[key].append(v)

        shutil.rmtree(tmp_dir)

    def _load_oc_template(self):
        """
        Function loads openshift template
        :return: YAML dictionary
        """
        if not self._exist_openshift_template():
            return None
        with open(self.oc_template, 'r') as f:
            try:
                templ = yaml.load(f)
            except yaml.YAMLError as exc:
                print(exc)
                raise
        return templ

    def _get_labels(self, templ):
        labels = None
        try:
            labels = templ['metadata']['labels']
        except KeyError:
            labels = {}
        raise_exception = False
        try:
            labels['description'] = self.docker_dict[LABEL]['description']
        except KeyError:
            print("Label Description is missing in Dockerfile. It is mandatory.")
            raise_exception = True
        try:
            labels['tags'] = self.docker_dict[LABEL]['io.openshift.tags']
        except KeyError:
            print('Label tags is missing in Dockerfile. It is mandatory.')
            raise_exception = True
        if raise_exception:
            raise KeyError
        labels['template'] = self.docker_image
        return labels

    def _get_docker_labels(self):
        """
        Function returns docker labels
        :return: label dictionary
        """
        if LABEL in self.docker_dict and self.docker_dict[LABEL]:
            return self.docker_dict[LABEL]
        return None

    def _get_docker_volumes(self):
        """
        Function returns docker volumes and labels
        :return: volume list, volume names
        """
        volume_list = []
        volume_names = []
        if VOLUME in self.docker_dict and self.docker_dict[VOLUME]:
            for p in self.docker_dict[VOLUME]:
                volume_list.append({'mountPath': p,
                                    'name': 'name' + p.replace('/', '-')})
                volume_names.append({'name': 'name' + p.replace('/', '-'),
                                     'emptyDir': {}
                                     })
        return volume_list, volume_names

    def _get_docker_env(self):
        """
        Function return docker ENV directives
        :return: list of ENV variables
        """
        env_list = []
        if ENV in self.docker_dict and self.docker_dict[ENV]:
            for e in self.docker_dict[ENV]:
                key, val = e.split('=')
                env_list.append({'name': key,
                                 'value': val})
        return env_list

    def _get_docker_expose(self):
        """
        Function return docker EXPOSE directives
        :return: list of PORTS
        """
        ports_list = []
        if EXPOSE in self.docker_dict and self.docker_dict[EXPOSE]:
            for p in self.docker_dict[EXPOSE]:
                ports_list.append({'containerPort': int(p)})
        return ports_list

    def write_oc_template(self, templ):
        """
        Function writes a YAML dictionary into template
        :param templ: YAML template with all data
        :return:
        """
        tmp_dir = tempfile.mkdtemp()
        if os.path.isdir(tmp_dir):
            shutil.rmtree(tmp_dir)
        os.makedirs(tmp_dir)
        tmp_file = os.path.join(tmp_dir, os.path.basename(self.oc_template))
        with open(tmp_file, 'w') as f:
            try:
                yaml.safe_dump(templ, f, default_flow_style=False)
                print("OpenShift template is generated here: %s" % (tmp_file))
            except yaml.YAMLError as exc:
                print(exc)
                raise

    def get_docker_directives(self, templ):
        """
        Function collects all directives
        :param templ:
        :return: label_list, volume_list, volume_names, env_list, ports_list
        """
        labels = volume_list = volume_names = env_list = ports_list = None
        if self.docker_dict:
            labels = self._get_labels(templ)
            volume_list, volume_names = self._get_docker_volumes()
            env_list = self._get_docker_env()
            ports_list = self._get_docker_expose()
        return labels, volume_list, volume_names, env_list, ports_list

    def generate_oc_template(self, templ, labels, volume_list, volume_names, env_list, ports_list):
        """
        Function fulfills template with data taken from Dockerfile.
        :param templ: YAML openshift templates
        :param labels: list of labels
        :param volume_list: volume list
        :param volume_names: volume names
        :param env_list: env list
        :param ports_list: port list
        :return: template with all data
        """
        templ['metadata']['name'] = self.docker_image
        templ['metadata']['labels'] = labels
        for obj in templ['objects']:
            obj['spec']['dockerImageRepository'] = self.docker_image
            obj['metadata']['name'] = self.docker_image
            if 'template' in obj['spec']:
                obj['spec']['template']['metadata']['labels']['name'] = self.docker_image
                containers = obj['spec']['template']['spec']['containers'][0]
                if env_list:
                    containers["env"] = env_list
                else:
                    containers.pop("env")
                if ports_list:
                    containers["ports"] = ports_list
                else:
                    containers.pop("ports")
                if volume_list:
                    containers['volumeMounts'] = volume_list
                    obj['spec']['template']['spec']['volumes'] = volume_names
                else:
                    containers.pop('volumeMounts')
                    obj['spec']['template']['spec'].pop('volumes')
                containers['name'] = self.docker_image
                containers['image'] = self.docker_image

            if 'triggers' in obj['spec']:
                for trig in obj['spec']['triggers']:
                    trig['imageChangeParams']['containerNames'] = [self.docker_image]
                    trig['imageChangeParams']['from']['name'] = self.docker_image + ":latest"

        return templ

    def run(self):
        """
        Main function
        :return:
        """
        self._get_openshift_template()
        if not self._exist_docker_file() or not self._exist_openshift_template():
            return 1
        self._get_docker_tags()
        templ = self._load_oc_template()
        try:
            tmpl = self.generate_oc_template(templ, *self.get_docker_directives(templ))
        except KeyError:
            return 1
        self.write_oc_template(tmpl)


